from abc import ABC, abstractmethod


class Spreadsheets(ABC):
    EXIT_CODE_INVALID_TOKEN = 200
    EXIT_CODE_INVALID_DATABASE_ID = 201
    EXIT_CODE_UPLOAD_ERROR = 202
    EXIT_CODE_DOWNLOAD_ERROR = 203
    EXIT_CODE_DB_CREATE_ERROR = 204
    EXIT_CODE_BAD_REQUEST = 205
    EXIT_CODE_FILE_NOT_FOUND = 206
    EXIT_CODE_UNKNOWN_ERROR = 249
    EXIT_CODE_RATE_LIMIT = 250
    EXIT_CODE_INVALID_INPUT = 251

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def upload(self):
        pass

    @abstractmethod
    def fetch(self):
        pass



import os
import logging


def add_logging_level(level_name, level_num, method_name=None):
    """
    Adds a new logging level to the logging module and the currently configured logging class.
    """
    if not method_name:
        method_name = level_name.lower()

    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    if not hasattr(logging, level_name):
        logging.addLevelName(level_num, level_name)
        setattr(logging, level_name, level_num)

    if not hasattr(logging.getLoggerClass(), method_name):
        setattr(logging.getLoggerClass(), method_name, log_for_level)

    if not hasattr(logging, method_name):
        setattr(logging, method_name, log_to_root)


class ShipyardLogger:
    _logger = None
    AUTHTEST_LEVEL = 100

    @classmethod
    def get_logger(cls):
        if cls._logger is None:
            cls._logger = logging.getLogger("Shipyard")
            log_level = os.getenv("LOG_LEVEL", "INFO").upper()

            add_logging_level("AUTHTEST", cls.AUTHTEST_LEVEL)

            console = logging.StreamHandler()
            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s -%(lineno)d: %(message)s"
            )
            console.setFormatter(formatter)
            cls._logger.addHandler(console)

            try:
                cls._logger.setLevel(log_level)
            except ValueError:
                cls._logger.setLevel(logging.INFO)
                cls._logger.warning(
                    f"Invalid log level {log_level}. Defaulting to INFO."
                )
            else:
                cls._logger.debug(f"Log level set to {log_level}")
        return cls._logger



import re
from abc import ABC, abstractmethod

FILE_PATTERN = re.compile(r"\{\{[^\{\}]+\}\}")


class Messaging(ABC):
    @abstractmethod
    def connect(self):
        pass

    EXIT_CODE_INVALID_CREDENTIALS = 200
    EXIT_CODE_INVALID_INPUT = 201
    EXIT_CODE_RATE_LIMIT = 202
    EXIT_CODE_BAD_REQUEST = 203
    EXIT_CODE_FILE_NOT_FOUND = 204
    EXIT_CODE_UNKNOWN_ERROR = 249

    @classmethod
    def message_content_file_injection(self, message: str) -> str:
        """
        Extracts the contents of a file from a message, if a file pattern is present.
        Message pattern: {{text:file.txt}}

        Args:
            message (str): The message

        Returns:
            str: The message with the file contents injected, if a file pattern is present.

        Raises:
            ExitCodeException: If the file pattern is invalid or the file is not found.
        """

        def _extract_filename(message: str) -> str:
            from shipyard_templates import ExitCodeException

            match = FILE_PATTERN.search(message)
            if not match:
                raise ExitCodeException(
                    "No file pattern found in the message.",
                    self.EXIT_CODE_INVALID_INPUT,
                )
            matched = match.group()[
                2:-2
            ].strip()  # Extract the matched text without the curly braces

            if not matched.startswith("text:"):
                raise ExitCodeException(
                    "When using the file pattern, the parameter needs to be prefixed with 'text:'",
                    self.EXIT_CODE_INVALID_INPUT,
                )
            return matched[5:]  # Remove the 'text:' prefix

        def _read_file(file_path: str, message: str) -> str:
            from shipyard_templates import ExitCodeException

            try:
                with open(file_path.strip(), "r") as f:
                    content = f.read()
            except FileNotFoundError as e:
                raise ExitCodeException(
                    f"Could not load the contents of file {file_path}. Make sure the file exists",
                    Messaging.EXIT_CODE_FILE_NOT_FOUND,
                ) from e
            except OSError as e:
                raise ExitCodeException(
                    f"Error opening file {file_path}.",
                    Messaging.EXIT_CODE_INVALID_INPUT,
                ) from e

            return FILE_PATTERN.sub("", message, count=1) + "\n\n" + content

        if not re.search(FILE_PATTERN, message):
            return message

        file = _extract_filename(message)
        return _read_file(file, message)



from abc import ABC, abstractmethod


class DigitalAdvertising(ABC):
    @abstractmethod
    def connect(self, **kwargs):
        pass

    @abstractmethod
    def create(self, **kwargs):
        pass

    @abstractmethod
    def read(self, **kwargs):
        pass

    @abstractmethod
    def update(self, **kwargs):
        pass

    @abstractmethod
    def delete(self, **kwargs):
        pass



from abc import ABC, abstractmethod


class CloudStorage(ABC):
    EXIT_CODE_INVALID_INPUT = 200
    EXIT_CODE_INVALID_CREDENTIALS = 202
    EXIT_CODE_FILE_ACCESS_ERROR = 203
    EXIT_CODE_FOLDER_ACCESS_ERROR = 204
    EXIT_CODE_UPLOAD_ERROR = 205
    EXIT_CODE_DOWNLOAD_ERROR = 206
    EXIT_CODE_FILE_NOT_FOUND = 207
    EXIT_CODE_FOLDER_CREATION_ERROR = 208
    EXIT_CODE_FILE_MATCH_ERROR = 210
    EXIT_CODE_UNKNOWN_ERROR = 249

    @abstractmethod
    def download(self):
        pass

    @abstractmethod
    def move(self):
        pass

    @abstractmethod
    def remove(self):
        pass

    @abstractmethod
    def upload(self):
        pass

    @abstractmethod
    def connect(self):
        pass



from abc import ABC, abstractmethod


class ExitCodeError(Exception):
    def __init__(self, message, exit_code):
        super().__init__(message)
        self.message = message
        self.exit_code = exit_code


class ProjectManagement(ABC):
    # Class level exit codes
    # general exit codes
    EXIT_CODE_INVALID_CREDENTIALS = 200
    EXIT_CODE_BAD_REQUEST = 201
    EXIT_CODE_TICKET_NOT_FOUND = 202
    EXIT_CODE_INVALID_ISSUE_TYPE = 203
    EXIT_CODE_INVALID_STATUS = 204
    EXIT_CODE_INVALID_INPUT = 205
    EXIT_CODE_UNKNOWN_ERROR = 299

    def __init__(self, access_token: str, **kwargs) -> None:
        self.access_token = access_token

    @abstractmethod
    def create_ticket(self, **kwargs):
        pass

    @abstractmethod
    def get_ticket(self, **kwargs):
        pass

    @abstractmethod
    def update_ticket(self, **kwargs):
        pass

    @abstractmethod
    def connect(self, **kwargs):
        pass



from abc import ABC, abstractmethod
from .exit_code_exception import ExitCodeException


class UnauthorizedError(ExitCodeException):
    def __init__(self, message: str):
        self.message = f"Failed to connect to the API. Ensure that the credential item provided is correct. Message from the API: {message}"
        self.exit_code = Etl.EXIT_CODE_INVALID_CREDENTIALS


class BadRequestError(ExitCodeException):
    def __init__(self, message: str):
        self.message = f"The request is malformed. Message from the API: {message}"
        self.exit_code = Etl.EXIT_CODE_BAD_REQUEST


from .exit_code_exception import ExitCodeException


class UnauthorizedError(ExitCodeException):
    def __init__(self, message: str):
        self.message = (
            f"Failed to connect to the API. Ensure that the credential item provided is correct. Message "
            f"from the API: {message}"
        )
        self.exit_code = Etl.EXIT_CODE_INVALID_CREDENTIALS


class BadRequestError(ExitCodeException):
    def __init__(self, message: str):
        self.message = f"The request is malformed. Message from the API: {message}"
        self.exit_code = Etl.EXIT_CODE_BAD_REQUEST


class UnknownError(ExitCodeException):
    def __init__(self, message: str):
        self.message = f"An unexpected error occurred. Message from the API: {message}"
        self.exit_code = Etl.EXIT_CODE_UNKNOWN_ERROR


class Etl(ABC):
    # Class level exit codes
    # general exit codes
    EXIT_CODE_FINAL_STATUS_COMPLETED = 0
    EXIT_CODE_INVALID_CREDENTIALS = 200
    EXIT_CODE_BAD_REQUEST = 201
    EXIT_CODE_INVALID_INPUT = 206
    # trigger sync exit codes
    EXIT_CODE_SYNC_REFRESH_ERROR = 202
    EXIT_CODE_SYNC_ALREADY_RUNNING = 203
    EXIT_CODE_SYNC_INVALID_SOURCE_ID = 204
    EXIT_CODE_SYNC_INVALID_POKE_INTERVAL = 205
    EXIT_CODE_SYNC_CHECK_ERROR = 220
    # verify status exit codes
    EXIT_CODE_FINAL_STATUS_INCOMPLETE = 210
    EXIT_CODE_FINAL_STATUS_ERRORED = 211
    EXIT_CODE_FINAL_STATUS_INVALID = 212
    EXIT_CODE_FINAL_STATUS_CANCELLED = 213
    EXIT_CODE_FINAL_STATUS_PENDING = 214
    EXIT_CODE_UNKNOWN_STATUS = 215
    EXIT_CODE_UNKNOWN_ERROR = 249

    # API TIMEOUT
    TIMEOUT = 30

    def __init__(self, access_token: str, **kwargs) -> None:
        self.access_token = access_token

    @abstractmethod
    def trigger_sync(self, **kwargs):
        pass

    @abstractmethod
    def determine_sync_status(self, **kwargs):
        pass



from abc import ABC, abstractmethod

class Presentations(ABC):
    EXIT_CODE_INVALID_TOKEN = 300
    EXIT_CODE_INVALID_PRESENTATION_ID = 301
    EXIT_CODE_UPLOAD_ERROR = 302
    EXIT_CODE_DOWNLOAD_ERROR = 303
    EXIT_CODE_CREATE_ERROR = 304
    EXIT_CODE_BAD_REQUEST = 305
    EXIT_CODE_FILE_NOT_FOUND = 306
    EXIT_CODE_UNKNOWN_ERROR = 349
    EXIT_CODE_RATE_LIMIT = 350
    EXIT_CODE_INVALID_INPUT = 351

    @abstractmethod
    def connect(self):
        """Establish connection to the Google Slides and Drive services."""
        pass

    @abstractmethod
    def create(self, title: str) -> str:
        """Create a new presentation and return the presentation ID."""
        pass

    @abstractmethod
    def upload(self, presentation_id: str, slides_data: list):
        """Add slides to the specified presentation."""
        pass

    @abstractmethod
    def share(self, presentation_id: str) -> str:
        """Make the presentation shareable and return the shareable link."""
        pass



from abc import ABC, abstractmethod


class DataVisualization(ABC):
    EXIT_CODE_INVALID_CREDENTIALS = 201
    EXIT_CODE_BAD_REQUEST = 202
    EXIT_CODE_UNKNOWN_ERROR = 203
    EXIT_CODE_RATE_LIMIT = 204
    EXIT_CODE_SERVICE_UNAVAILABLE = 205
    EXIT_CODE_INVALID_INPUT = 206

    @abstractmethod
    def connect(self, **kwargs):
        pass



from abc import ABC, abstractmethod


class Crm(ABC):
    # Class level exit codes
    EXIT_CODE_EXPORT_NOT_FINISHED = 101
    EXIT_CODE_FILE_NOT_FOUND = 102
    EXIT_CODE_MULTIPLE_RECORDS_FOUND = 103
    EXIT_CODE_NOT_MODIFIED = 104
    EXIT_CODE_CONFLICT = 105
    EXIT_CODE_RESOURCE_NOT_FOUND = 106
    EXIT_CODE_UPLOAD_FAILED = 107

    # GENERIC EXIT CODES
    EXIT_CODE_INVALID_CREDENTIALS = 201
    EXIT_CODE_BAD_REQUEST = 202
    EXIT_CODE_UNKNOWN_ERROR = 203
    EXIT_CODE_RATE_LIMIT = 204
    EXIT_CODE_SERVICE_UNAVAILABLE = 205
    EXIT_CODE_INVALID_INPUT = 206
    # API TIMEOUT
    TIMEOUT = 30

    def __init__(self, access_token: str, **kwargs) -> None:
        self.access_token = access_token

    @abstractmethod
    def connect(self, **kwargs):
        pass

    @abstractmethod
    def export_data(self, **kwargs):
        pass

    @abstractmethod
    def import_data(self, **kwargs):
        pass



from abc import ABC, abstractmethod
from .shipyard_logger import ShipyardLogger
from .exit_code_exception import ExitCodeException
from typing import Optional


class UploadError(ExitCodeException):
    def __init__(self, table: str, error_msg: Exception):
        self.message = (
            f"Error in loading data to {table}. Message from the server is: {error_msg}"
        )
        self.exit_code = Database.EXIT_CODE_UPLOAD_ERROR


class FetchError(ExitCodeException):
    def __init__(self, error_msg: Exception):
        self.message = (
            f"Error in downloading data. Message from the server is: {error_msg}"
        )
        self.exit_code = Database.EXIT_CODE_DOWNLOAD_ERROR


class QueryError(ExitCodeException):
    def __init__(self, error_msg: Exception):
        self.message = (
            f"Error in executing query. Message from the server is: {error_msg}"
        )
        self.exit_code = Database.EXIT_CODE_QUERY_ERROR


class ConnectionError(ExitCodeException):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = Database.EXIT_CODE_INVALID_CREDENTIALS


class Database(ABC):
    EXIT_CODE_INVALID_CREDENTIALS = 200
    EXIT_CODE_INVALID_ACCOUNT = 201  # snowflake specific
    EXIT_CODE_INVALID_WAREHOUSE = 202
    EXIT_CODE_INVALID_DATABASE = 203
    EXIT_CODE_INVALID_SCHEMA = 204
    EXIT_CODE_INVALID_QUERY = 205
    EXIT_CODE_NO_RESULTS = 206
    EXIT_CODE_INVALID_UPLOAD_COLUMNS = 207
    EXIT_CODE_INVALID_ARGUMENTS = 208
    EXIT_CODE_INVALID_DATA_TYPES = 209
    EXIT_CODE_FILE_NOT_FOUND = 210
    EXIT_CODE_INVALID_UPLOAD_VALUE = 211
    EXIT_CODE_NO_FILE_MATCHES = 212

    # method errors
    EXIT_CODE_UPLOAD_ERROR = 220
    EXIT_CODE_QUERY_ERROR = 221
    EXIT_CODE_DOWNLOAD_ERROR = 222

    EXIT_CODE_UNKNOWN = 249

    def __init__(self, username: str, password: Optional[str] = None, **kwargs) -> None:
        self.username = username
        self.password = password

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def execute_query(self, **kwargs):
        pass

    @abstractmethod
    def fetch(self, **kwargs):
        pass

    @abstractmethod
    def upload(self, **kwargs):
        pass


class GoogleDatabase:
    EXIT_CODE_INVALID_CREDENTIALS = 200
    EXIT_CODE_INVALID_ACCOUNT = 201  # snowflake specific
    EXIT_CODE_INVALID_WAREHOUSE = 202
    EXIT_CODE_INVALID_DATABASE = 203
    EXIT_CODE_INVALID_SCHEMA = 204  # snowflake specific
    EXIT_CODE_INVALID_QUERY = 205
    EXIT_CODE_NO_RESULTS = 206
    EXIT_CODE_FILE_NOT_FOUND = 205
    EXIT_CODE_INVALID_UPLOAD_VALUE = 206
    EXIT_CODE_INVALID_UPLOAD_COLUMNS = 207
    EXIT_CODE_INVALID_ARGUMENTS = 208
    EXIT_CODE_INVALID_DATA_TYPES = 209

    def __init__(self, service_account: str, **kwargs) -> None:
        self.service_account = service_account

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def execute_query(self, **kwargs):
        pass

    @abstractmethod
    def fetch(self, **kwargs):
        pass

    @abstractmethod
    def upload(self, **kwargs):
        pass


class DatabricksDatabase:
    EXIT_CODE_INVALID_CREDENTIALS = 200
    EXIT_CODE_INVALID_WAREHOUSE = 202
    EXIT_CODE_INVALID_DATABASE = 203
    EXIT_CODE_INVALID_SCHEMA = 204
    EXIT_CODE_INVALID_QUERY = 205
    EXIT_CODE_NO_RESULTS = 206
    EXIT_CODE_FILE_NOT_FOUND = 205
    EXIT_CODE_INVALID_UPLOAD_VALUE = 206
    EXIT_CODE_INVALID_UPLOAD_COLUMNS = 207
    EXIT_CODE_INVALID_ARGUMENTS = 208
    EXIT_CODE_INVALID_DATA_TYPES = 209
    EXIT_CODE_INSERT_STATEMENT_ERROR = 210
    EXIT_CODE_CREATE_STATEMENT_ERROR = 211
    EXIT_CODE_UNKNOWN_ERROR = 249

    def __init__(
        self, server_host: str, http_path: str, access_token: str, **kwargs
    ) -> None:
        self.server_host = server_host
        self.http_path = http_path
        self.access_token = access_token

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def execute_query(self, **kwargs):
        pass

    @abstractmethod
    def fetch(self, **kwargs):
        pass

    @abstractmethod
    def upload(self, **kwargs):
        pass



from .exit_code_exception import ExitCodeException
from typing import Union

EXIT_CODE_INVALID_CREDENTIAL = 180
EXIT_CODE_BAD_API_REQUEST = 181
EXIT_CODE_ACCESS_DENIED = 182
EXIT_CODE_NOT_FOUND = 183
EXIT_CODE_INTERNAL_SERVER_ERROR = 184
EXIT_CODE_UNKNOWN_ERROR = 249


class InvalidCredentialError(ExitCodeException):
    def __init__(self, message):
        super().__init__(message, exit_code=EXIT_CODE_INVALID_CREDENTIAL)


class BadRequestError(ExitCodeException):
    def __init__(self, message):
        super().__init__(message, exit_code=EXIT_CODE_BAD_API_REQUEST)


class AccessDeniedError(ExitCodeException):
    def __init__(self, message):
        super().__init__(message, exit_code=EXIT_CODE_ACCESS_DENIED)


class NotFoundError(ExitCodeException):
    def __init__(self, message):
        super().__init__(message, exit_code=EXIT_CODE_NOT_FOUND)


class InternalServerError(ExitCodeException):
    def __init__(self, message):
        super().__init__(message, exit_code=EXIT_CODE_INTERNAL_SERVER_ERROR)


def handle_errors(error_message: str, error_code: Union[int, str]):
    """Helper function to raise the appropriate error based on the HTTP status code

    Args:
        error_message: The message returned by the server
        error_code: The HTTP status code

    Raises:
        InvalidCredentialError:
        BadRequestError:
        AccessDeniedError:
        NotFoundError:
        InternalServerError:
        ExitCodeException:
    """
    if error_code == 401:
        raise InvalidCredentialError(
            f"Invalid credentials. Response from the server: {error_message}"
        )
    elif error_code == 400:
        raise BadRequestError(f"Bad request. Response from the server: {error_message}")
    elif error_code == 403:
        raise AccessDeniedError(
            f"Access denied. Response from the server: {error_message}"
        )
    elif error_code == 404:
        raise NotFoundError(
            f"Resource not found. Response from the server: {error_message}"
        )
    elif error_code == 500:
        raise InternalServerError(
            f"Internal server error. Response from the server: {error_message}"
        )
    else:
        raise ExitCodeException(
            f"An unexpected error occurred: {error_message}",
            exit_code=EXIT_CODE_UNKNOWN_ERROR,
        )



def standardize_errors(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ExitCodeException as error:
            raise ExitCodeException(error.message, error.exit_code) from error
        except Exception as error:
            raise ExitCodeException(
                f"Unknown Error occurred: {str(error)}", 1
            ) from error

    return wrapper


class ExitCodeException(Exception):
    def __init__(self, message, exit_code):
        super().__init__(message)
        self.message = message
        self.exit_code = exit_code



from .cloudstorage import CloudStorage
from .database import Database, GoogleDatabase, DatabricksDatabase
from .datavisualization import DataVisualization
from .etl import Etl
from .messaging import Messaging
from .notebooks import Notebooks
from .shipyard_logger import ShipyardLogger
from .spreadsheets import Spreadsheets
from .presentations import Presentations
from .projectmanagement import ProjectManagement, ExitCodeError
from .exit_code_exception import ExitCodeException, standardize_errors
from .crm import Crm
from .digital_advertising import DigitalAdvertising
from .errors import (
    InvalidCredentialError,
    BadRequestError,
    AccessDeniedError,
    NotFoundError,
    InternalServerError,
    handle_errors,
)
